use axum::extract::State;
use axum::{routing, Router};
use axum_template::{engine::Engine, Key, RenderHtml};
use minijinja::{path_loader, Environment};
use state::{AppState, TemplateEngine};
use std::net::SocketAddr;

use log::info;
use serde::Serialize;
use std::env;
use std::sync::{Arc, Mutex};

use dotenvy::dotenv;
use vbplay::{AudioThread, DeviceSelection, SoundDecoder};

mod soundclips;
mod state;
mod vbplay;

#[tokio::main]
async fn main() {
    dotenv().ok();

    let device_pattern = env::var("DEVICE_PATTERN").expect("Need a device pattern");
    let folder = env::var("DIRECTORY").expect("Need a directory in env file.");

    let mut jinja = Environment::new();
    jinja.set_loader(path_loader("templates"));
    let template_engine = Engine::from(jinja);

    let audio = vbplay::audio_thread(
        DeviceSelection::FindByPattern(device_pattern),
        SoundDecoder::Detect,
    );
    let audio: Arc<Mutex<AudioThread>> = Arc::new(Mutex::new(audio));

    let app_state = AppState {
        engine: template_engine,
        clips: soundclips::scan_directory_for_clips(&folder, &["mp3", "ogg", "wav", "flac"])
            .expect("No Soundclips found."),
        player: audio,
        playback: None,
    };

    // Set the address to run our application on
    let addr = SocketAddr::from(([0, 0, 0, 0], 3000));

    // Build our application with a route
    let app = Router::new()
        .route("/", routing::get(home))
        .route("/play/:hash", routing::get(play_handler))
        .route("/stop", routing::get(stop_handler))
        .with_state(app_state);

    println!("listening on {}", addr);

    // Run it with hyper
    axum::Server::bind(&addr)
        .serve(app.into_make_service())
        .await
        .unwrap();
}

#[derive(Serialize)]
struct TemplateContext {
    clips: Vec<soundclips::SoundClip>,
}

async fn home(engine: TemplateEngine, state: State<AppState>) -> impl axum::response::IntoResponse {
    let clips = state.0.clone().clips; // this is not ideal.
    let context = TemplateContext { clips: clips };
    RenderHtml(Key("soundboard.jinja".to_owned()), engine, context)
}

async fn play_handler(
    axum::extract::Path(hash): axum::extract::Path<String>,
    state: State<AppState>,
) -> String {
    if let Some(clip) = state
        .clips
        .clone()
        .into_iter()
        .find(|s| hash == s.hash().to_string())
    {
        let player = state.player.clone();
        let full_file_name = &clip.full_file_name();
        info!("Playing {}", full_file_name);
        player
            .lock()
            .unwrap()
            .send(vbplay::Command::PlayWhilePressing(
                full_file_name.into(),
                "".to_owned(), // placeholder.
            ))
            .unwrap();
    }
    "".to_owned()
}

async fn stop_handler(state: State<AppState>) -> String {
    let player = state.player.clone();
    player.lock().unwrap().send(vbplay::Command::Stop).unwrap();

    "".to_owned()
}
